import importlib

import plotly.express as plotly_express
import plotly.graph_objects as go
import pytest

try:
    from pydantic.v1 import Field, ValidationError
except ImportError:  # pragma: no cov
    from pydantic import Field, ValidationError

from vizro.charts._charts_utils import _DashboardReadyFigure
from vizro.models import VizroBaseModel
from vizro.models.types import CapturedCallable, capture


def positional_only_function(a, /):
    pass


def var_positional_function(*args):
    pass


@pytest.mark.parametrize("function", [positional_only_function, var_positional_function])
def test_invalid_parameter_kind(function):
    with pytest.raises(
        ValueError,
        match="CapturedCallable does not accept functions with positional-only or variadic positional parameters",
    ):
        CapturedCallable(function)


def positional_or_keyword_function(a, b, c):
    return a + b + c


def keyword_only_function(a, *, b, c):
    return a + b + c


def var_keyword_function(a, **kwargs):
    return a + kwargs["b"] + kwargs["c"]


@pytest.fixture
def captured_callable(request):
    return CapturedCallable(request.param, 1, b=2)


@pytest.mark.parametrize(
    "captured_callable",
    [positional_or_keyword_function, keyword_only_function, var_keyword_function],
    indirect=True,
)
class TestCallKeywordArguments:
    def test_call_provide_required_argument(self, captured_callable):
        assert captured_callable(c=3) == 1 + 2 + 3

    def test_call_override_existing_arguments(self, captured_callable):
        assert captured_callable(a=5, b=2, c=6) == 5 + 2 + 6


@pytest.mark.parametrize(
    "captured_callable",
    [positional_or_keyword_function],
    indirect=True,
)
def test_call_positional_arguments(captured_callable):
    assert captured_callable(3) == 1 + 2 + 3


@pytest.mark.parametrize(
    "captured_callable",
    [positional_or_keyword_function, keyword_only_function, var_keyword_function],
    indirect=True,
)
class TestDunderMethods:
    def test_getitem_known_args(self, captured_callable):
        assert captured_callable["a"] == 1
        assert captured_callable["b"] == 2

    def test_getitem_unknown_args(self, captured_callable):
        with pytest.raises(KeyError):
            captured_callable["c"]

    def test_delitem(self, captured_callable):
        del captured_callable["a"]

        with pytest.raises(KeyError):
            captured_callable["a"]


@pytest.mark.parametrize(
    "captured_callable",
    [positional_or_keyword_function, keyword_only_function, var_keyword_function],
    indirect=True,
)
def test_call_positional_and_keyword_supplied(captured_callable):
    with pytest.raises(ValueError, match="does not support calling with both positional and keyword arguments"):
        captured_callable(3, c=4)


@pytest.mark.parametrize(
    "captured_callable, expectation",
    [
        (positional_or_keyword_function, pytest.raises(TypeError, match="missing 1 required positional argument: 'c'")),
        (keyword_only_function, pytest.raises(TypeError, match="missing 1 required keyword-only argument: 'c'")),
        (var_keyword_function, pytest.raises(KeyError, match="'c'")),
    ],
    indirect=["captured_callable"],
)
class TestCallMissingArgument:
    def test_call_missing_argument(self, captured_callable, expectation):
        with expectation:
            captured_callable()

    def test_call_is_memoryless(self, captured_callable, expectation):
        captured_callable(c=3)

        with expectation:
            captured_callable()


@pytest.mark.parametrize(
    "captured_callable, expectation",
    [
        (
            positional_or_keyword_function,
            pytest.raises(TypeError, match="takes 1 positional arguments but 2 were given"),
        ),
        (keyword_only_function, pytest.raises(TypeError, match="takes 0 positional arguments but 2 were given")),
        (var_keyword_function, pytest.raises(TypeError, match="takes 0 positional arguments but 2 were given")),
    ],
    indirect=["captured_callable"],
)
class TestCallPositionalArguments:
    def test_call_positional_arguments_invalid(self, captured_callable, expectation):
        with expectation:
            captured_callable(3, 4)


@pytest.mark.parametrize(
    "captured_callable, expectation",
    [
        (positional_or_keyword_function, pytest.raises(TypeError, match="got an unexpected keyword argument")),
        (keyword_only_function, pytest.raises(TypeError, match="got an unexpected keyword argument")),
        (var_keyword_function, pytest.raises(KeyError, match="'c'")),
    ],
    indirect=["captured_callable"],
)
def test_call_unknown_argument(captured_callable, expectation):
    with expectation:
        captured_callable(e=1)


def undecorated_function(a, b, c, d=4):
    return a + b + c + d


@capture("action")
def decorated_action_function(a, b, c, d=4):
    return a + b + c + d


@capture("graph")
def decorated_graph_function(data_frame):
    return go.Figure()


@capture("graph")
def decorated_graph_function_px(data_frame):
    return plotly_express.scatter(
        data_frame=data_frame,
        x="gdpPercap",
        y="lifeExp",
    )


@capture("graph")
def invalid_decorated_graph_function():
    return go.Figure()


class Model(VizroBaseModel):
    # The import_path here makes it possible to import the above function using getattr(import_path, _target_).
    function: CapturedCallable = Field(..., import_path=importlib.import_module(__name__))


class TestModelFieldPython:
    def test_decorated_action_function(self):
        model = Model(function=decorated_action_function(a=1, b=2))
        assert model.function(c=3, d=4) == 1 + 2 + 3 + 4

    def test_decorated_graph_function(self):
        model = Model(function=decorated_graph_function(data_frame=None))
        assert model.function() == go.Figure()

    def test_undecorated_function(self):
        with pytest.raises(ValidationError, match="provide a valid CapturedCallable object"):
            Model(function=undecorated_function(1, 2, 3))


class TestCapture:
    def test_decorated_df_standard_case(self, gapminder):
        fig = decorated_graph_function_px(gapminder)
        assert len(fig.data) > 0
        assert fig.__class__ == _DashboardReadyFigure

    def test_decorated_df_str(self):
        fig = decorated_graph_function_px("gapminder")
        assert fig == _DashboardReadyFigure()
        assert len(fig.data) == 0

    def test_decorated_graph_function_missing_data_frame(self):
        with pytest.raises(
            ValueError,
            match="decorated_graph_function must supply a value to data_frame argument",
        ):
            decorated_graph_function()

    def test_invalid_decorated_graph_function(self):
        with pytest.raises(
            ValueError,
            match="invalid_decorated_graph_function must have data_frame argument",
        ):
            invalid_decorated_graph_function()


class TestModelFieldJSONConfig:
    def test_no_arguments(self):
        config = {"_target_": "decorated_action_function"}
        model = Model(function=config)
        assert isinstance(model.function, CapturedCallable)
        assert model.function(a=1, b=2, c=3, d=4) == 1 + 2 + 3 + 4

    def test_some_arguments(self):
        config = {"_target_": "decorated_action_function", "a": 1, "b": 2}
        model = Model(function=config)
        assert isinstance(model.function, CapturedCallable)
        assert model.function(c=3) == 1 + 2 + 3 + 4

    def test_all_arguments(self):
        config = {"_target_": "decorated_action_function", "a": 1, "b": 2, "c": 3}
        model = Model(function=config)
        assert isinstance(model.function, CapturedCallable)
        assert model.function() == 1 + 2 + 3 + 4

    def test_decorated_graph_function(self):
        config = {"_target_": "decorated_graph_function", "data_frame": None}
        model = Model(function=config)
        assert model.function() == go.Figure()

    def test_no_target(self):
        config = {"a": 1, "b": 2}
        with pytest.raises(ValidationError, match="must contain the key '_target_'"):
            Model(function=config)

    def test_invalid_import(self):
        config = {"_target_": "invalid_function"}
        with pytest.raises(ValidationError, match="_target_=invalid_function cannot be imported"):
            Model(function=config)

    def test_invalid_arguments(self):
        config = {"_target_": "decorated_action_function", "e": 5}
        with pytest.raises(ValidationError, match="got an unexpected keyword argument"):
            Model(function=config)

    def test_undecorated_function(self):
        config = {"_target_": "undecorated_function", "a": 1, "b": 2, "c": 3}
        with pytest.raises(ValidationError, match="_target_=undecorated_function must be wrapped"):
            Model(function=config)



"""Unit tests for vizro.models.Checklist."""
import json

import plotly
import pytest
from dash import dcc, html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

from vizro.models._action._action import Action
from vizro.models._components.form import Checklist


@pytest.fixture()
def expected_checklist():
    return html.Div(
        [
            html.P("Title"),
            dcc.Checklist(
                id="checklist_id",
                options=["ALL", "A", "B", "C"],
                value=["ALL"],
                className="selector_body_checklist",
                persistence=True,
                persistence_type="session",
            ),
        ],
        className="selector_container",
        id="checklist_id_outer",
    )


class TestChecklistInstantiation:
    """Tests model instantiation."""

    def test_create_checklist_mandatory_only(self):
        checklist = Checklist()

        assert hasattr(checklist, "id")
        assert checklist.type == "checklist"
        assert checklist.options == []
        assert checklist.value is None
        assert checklist.title == ""
        assert checklist.actions == []

    def test_create_checklist_mandatory_and_optional(self):
        checklist = Checklist(options=["A", "B", "C"], value=["A"], title="Title", id="checklist-id")

        assert checklist.id == "checklist-id"
        assert checklist.type == "checklist"
        assert checklist.options == ["A", "B", "C"]
        assert checklist.value == ["A"]
        assert checklist.title == "Title"
        assert checklist.actions == []

    @pytest.mark.parametrize(
        "test_options, expected",
        [
            (["A", "B", "C"], ["A", "B", "C"]),
            ([1, 2, 3], [1.0, 2.0, 3.0]),
            ([1.0, 2.0, 3.0], [1.0, 2.0, 3.0]),
            ([True, False, True], [True, False, True]),
            (
                [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}],
                [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}],
            ),
            (
                [{"label": "New York", "value": "NYC"}, {"label": "Berlin", "value": "BER"}],
                [{"label": "New York", "value": "NYC"}, {"label": "Berlin", "value": "BER"}],
            ),
            (
                [{"label": "True", "value": True}, {"label": "False", "value": False}],
                [{"label": "True", "value": True}, {"label": "False", "value": False}],
            ),
            ([True, 2.0, 1.0, "A", "B"], ["True", "2.0", "1.0", "A", "B"]),
        ],
    )
    def test_create_checklist_valid_options(self, test_options, expected):
        checklist = Checklist(options=test_options)

        assert hasattr(checklist, "id")
        assert checklist.type == "checklist"
        assert checklist.options == expected
        assert checklist.value is None
        assert checklist.title == ""
        assert checklist.actions == []

    @pytest.mark.parametrize("test_options", [1, "A", True, 1.0])
    def test_create_checklist_invalid_options_type(self, test_options):
        with pytest.raises(ValidationError, match="value is not a valid list"):
            Checklist(options=test_options)

    def test_create_checklist_invalid_options_dict(self):
        with pytest.raises(
            ValidationError,
            match="Invalid argument `options` passed. Expected a dict with keys `label` and `value`.",
        ):
            Checklist(options=[{"hello": "A", "world": "A"}, {"hello": "B", "world": "B"}])

    @pytest.mark.parametrize(
        "test_value, options",
        [
            (["A"], ["A", "B", "C"]),
            ([1, 2], [1, 2, 3]),
            ([1.0, 2.0], [1.0, 2.0, 3.0]),
            ([False, True], [True, False]),
            (["A", "B"], [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}]),
            (["True", "A"], [True, 2.0, 1.0, "A", "B"]),
        ],
    )
    def test_create_checklist_valid_value(self, test_value, options):
        checklist = Checklist(options=options, value=test_value)

        assert hasattr(checklist, "id")
        assert checklist.type == "checklist"
        assert checklist.value == test_value
        assert checklist.title == ""
        assert checklist.actions == []

    @pytest.mark.parametrize(
        "test_value, options",
        [
            ([5], ["A", "B", "C"]),
            ([5], [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}]),
            (["D"], ["A", "B", "C"]),
            (["D"], [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}]),
            ([True], ["A", "B", "C"]),
            ([True], [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}]),
        ],
    )
    def test_create_checklist_invalid_value_non_existing(self, test_value, options):
        with pytest.raises(ValidationError, match="Please provide a valid value from `options`."):
            Checklist(value=test_value, options=options)

    def test_create_checklist_invalid_value_format(self):
        with pytest.raises(ValidationError, match="value is not a valid list"):
            Checklist(value="A", options=["A", "B", "C"])

    def test_set_action_via_validator(self, identity_action_function):
        checklist = Checklist(actions=[Action(function=identity_action_function())])
        actions_chain = checklist.actions[0]
        assert actions_chain.trigger.component_property == "value"


class TestChecklistBuild:
    """Tests model build method."""

    def test_checklist_build(self, expected_checklist):
        checklist = Checklist(options=["A", "B", "C"], id="checklist_id", title="Title").build()

        result = json.loads(json.dumps(checklist, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_checklist, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected



from __future__ import annotations

from typing import TYPE_CHECKING, List, Literal

from dash import html

try:
    from pydantic.v1 import validator
except ImportError:  # pragma: no cov
    from pydantic import validator

from vizro.models import VizroBaseModel
from vizro.models._components.form import (
    Button,
    Checklist,
    Dropdown,
    RadioItems,
    RangeSlider,
    Slider,
)
from vizro.models._models_utils import _log_call, get_unique_grid_component_ids
from vizro.models.types import _FormComponentType

if TYPE_CHECKING:
    from vizro.models import Layout


class Form(VizroBaseModel):
    """Container for all form components to be provided to [`Page`][vizro.models.Page].

    Args:
        type (Literal["form"]): Defaults to `"form"`.
        components (List[FormComponentType]): List of components used in the form.
        layout (Layout): Defaults to `None`.
    """

    type: Literal["form"] = "form"
    components: List[_FormComponentType]
    layout: Layout = None  # type: ignore[assignment]

    @validator("layout", always=True)
    def set_layout(cls, layout, values):
        from vizro.models import Layout

        if "components" not in values:
            return layout

        if layout is None:
            grid = [[i] for i in range(len(values["components"]))]
            return Layout(grid=grid)

        unique_grid_idx = get_unique_grid_component_ids(layout.grid)
        if len(unique_grid_idx) != len(values["components"]):
            raise ValueError("Number of form and grid components need to be the same.")
        return layout

    @_log_call
    def pre_build(self):
        for component in self.components:
            if isinstance(component, (Slider, RangeSlider)):
                if component.min is None or component.max is None:
                    raise TypeError(f"{component.type} requires the arguments 'min' and 'max' when used within Form.")

            if isinstance(component, (Checklist, Dropdown, RadioItems)) and not component.options:
                raise TypeError(f"{component.type} requires the argument 'options' when used within Form.")

    @_log_call
    def build(self):
        component_container = [
            html.Div(
                component.build(),
                style={
                    "gridColumn": f"{grid_coord.col_start}/{grid_coord.col_end}",
                    "gridRow": f"{grid_coord.row_start}/{grid_coord.row_end}",
                },
            )
            for component, grid_coord in zip(self.components, self.layout.component_grid_lines)
        ]
        return self._make_form_layout(component_container)

    def _make_form_layout(self, component_container):
        return html.Div(
            component_container,
            style={
                "gridRowGap": self.layout.row_gap,
                "gridColumnGap": self.layout.col_gap,
                "gridTemplateColumns": f"repeat({len(self.layout.grid[0])}, minmax({self.layout.col_min_width}, 1fr))",
                "gridTemplateRows": f"repeat({len(self.layout.grid)}, minmax({self.layout.row_min_height}, 1fr))",
            },
            className="component_container_grid",
            id=self.id,
        )


if __name__ == "__main__":
    from vizro.models import Layout

    Form.update_forward_refs(Layout=Layout)

    print(  # noqa: T201
        repr(
            Form(
                layout=Layout(grid=[[i] for i in range(7)], row_min_height="200px"),
                components=[
                    Checklist(options=["Option 1", "Option 2", "Option 3"]),
                    Dropdown(options=["Option 1", "Option 2", "Option 3"]),
                    RadioItems(options=["Option 1", "Option 2", "Option 3"]),
                    Slider(min=0, max=5, step=1),
                    RangeSlider(min=0, max=5, step=1),
                    Button(id="form"),
                ],
            )
        )
    )



# Keep this import at the top to avoid circular imports since it's used in every model.
from ._base import VizroBaseModel  # noqa: I001
from ._action import Action
from ._components import Card, Graph, Table
from ._components.form import Button, Checklist, Dropdown, RadioItems, RangeSlider, Slider
from ._controls import Filter, Parameter
from ._navigation.accordion import Accordion
from ._navigation.navigation import Navigation
from ._navigation.nav_bar import NavBar
from ._navigation.nav_link import NavLink
from ._dashboard import Dashboard
from ._layout import Layout
from ._page import Page

Page.update_forward_refs(
    Button=Button, Card=Card, Filter=Filter, Graph=Graph, Parameter=Parameter, Table=Table, Accordion=Accordion
)
Navigation.update_forward_refs(Accordion=Accordion, NavBar=NavBar, NavLink=NavLink)
Dashboard.update_forward_refs(Page=Page, Navigation=Navigation)
NavBar.update_forward_refs(NavLink=NavLink)
NavLink.update_forward_refs(Accordion=Accordion)

# Please keep alphabetically ordered
__all__ = [
    "Accordion",
    "Action",
    "Button",
    "Card",
    "Checklist",
    "Dashboard",
    "Dropdown",
    "Filter",
    "Graph",
    "Layout",
    "NavBar",
    "NavLink",
    "Navigation",
    "Page",
    "Parameter",
    "RadioItems",
    "RangeSlider",
    "Slider",
    "Table",
    "VizroBaseModel",
]



from vizro.models._components.button import Button
from vizro.models._components.form.checklist import Checklist
from vizro.models._components.form.dropdown import Dropdown
from vizro.models._components.form.radio_items import RadioItems
from vizro.models._components.form.range_slider import RangeSlider
from vizro.models._components.form.slider import Slider

# Please keep alphabetically ordered
__all__ = [
    "Button",
    "Checklist",
    "Dropdown",
    "RadioItems",
    "RangeSlider",
    "Slider",
]



"""Contains utilities to create required components for the action loop."""

from dash import dcc, html

from vizro.actions._action_loop._action_loop_utils import (
    _get_actions_chains_on_registered_pages,
    _get_actions_on_registered_pages,
)


def _get_action_loop_components() -> html.Div:
    """Gets all required components for the action loop.

    Returns:
        List of dcc or html components.
    """
    actions_chains = _get_actions_chains_on_registered_pages()
    actions = _get_actions_on_registered_pages()

    if not actions_chains:
        return html.Div(id="action_loop_components_div")

    # Fundamental components required for the smooth operation of the action loop mechanism.
    components = [
        dcc.Store(id="action_finished"),
        dcc.Store(id="remaining_actions", data=[]),
        html.Div(id="cycle_breaker_div", hidden=True),
        dcc.Store(id="cycle_breaker_empty_output_store"),
    ]

    # Additional component for every ActionChain in the system.
    # Represents a proxy component between visible UI component and the gateway of the action loop mechanism.
    # Required to avoid the "Unknown callback Input" issue for multiple page app examples.
    components.extend(
        [
            dcc.Store(
                id={"type": "gateway_input", "trigger_id": actions_chain.id},
                data=f"{actions_chain.id}",
            )
            for actions_chain in actions_chains
        ]
    )

    # Additional component for every Action in the system.
    # This component is injected as the only Input (trigger) inside each Action.
    # It enables that the action can be triggered only from the action loop mechanism.
    components.extend([dcc.Store(id={"type": "action_trigger", "action_name": action.id}) for action in actions])

    # Additional store with all action_triggers ids.
    components.append(
        dcc.Store(
            id="action_trigger_actions_id",
            data=[action.id for action in actions],
        )
    )

    # Additional store that maps the actions chain trigger id and the list of action ids that should be executed.
    components.append(
        dcc.Store(
            id="trigger_to_actions_chain_mapper",
            data={
                actions_chain.id: [action.id for action in actions_chain.actions] for actions_chain in actions_chains
            },
        )
    )

    return html.Div(components, id="action_loop_components_div")



from vizro.actions._filter_action import _filter
from vizro.actions._on_page_load_action import _on_page_load
from vizro.actions._parameter_action import _parameter
from vizro.actions.export_data_action import export_data
from vizro.actions.filter_interaction_action import filter_interaction

# Please keep alphabetically ordered
__all__ = ["export_data", "filter_interaction"]



import pytest

import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro


@pytest.fixture
def gapminder_2007(gapminder):
    return gapminder.query("year == 2007")


@pytest.fixture
def box_params():
    return {"x": "continent", "y": "lifeExp", "custom_data": ["continent"]}


@pytest.fixture
def box_chart(gapminder_2007, box_params):
    return px.box(gapminder_2007, **box_params).update_layout(margin_t=24)


@pytest.fixture
def scatter_params():
    return {"x": "gdpPercap", "y": "lifeExp"}


@pytest.fixture
def scatter_chart(gapminder_2007, scatter_params):
    return px.scatter(gapminder_2007, **scatter_params).update_layout(margin_t=24)


@pytest.fixture
def target_scatter_filtered_continent(request, gapminder_2007, scatter_params):
    continent = request.param
    data = gapminder_2007[gapminder_2007["continent"].isin(continent)]
    return px.scatter(data, **scatter_params).update_layout(margin_t=24)


@pytest.fixture
def target_box_filtered_continent(request, gapminder_2007, box_params):
    continent = request.param
    data = gapminder_2007[gapminder_2007["continent"].isin(continent)]
    return px.box(data, **box_params).update_layout(margin_t=24)


@pytest.fixture
def managers_one_page_two_graphs_one_button(box_chart, scatter_chart):
    """Instantiates a simple model_manager and data_manager with a page, two graph models and the button component."""
    vm.Page(
        id="test_page",
        title="My first dashboard",
        components=[
            vm.Graph(id="box_chart", figure=box_chart),
            vm.Graph(id="scatter_chart", figure=scatter_chart),
            vm.Button(id="button"),
        ],
    )
    Vizro._pre_build()


@pytest.fixture
def managers_one_page_two_graphs_one_table_one_button(box_chart, scatter_chart, dash_data_table_with_id):
    """Instantiates a simple model_manager and data_manager with a page, two graph models and the button component."""
    vm.Page(
        id="test_page",
        title="My first dashboard",
        components=[
            vm.Graph(id="box_chart", figure=box_chart),
            vm.Graph(id="scatter_chart", figure=scatter_chart),
            vm.Table(id="vizro_table", figure=dash_data_table_with_id),
            vm.Button(id="button"),
        ],
    )
    Vizro._pre_build()



"""Demo to show how to use asserts. These are not real tests that are run as part of testing, just a teaching aid."""
from typing import List

from asserts import STRIP_ALL, assert_component_equal
from dash import html

from vizro.models import VizroBaseModel


class X(VizroBaseModel):
    # Low-level contents model.
    text: str

    def build(self):
        return html.Div(
            [html.H1("Heading"), html.P(self.text, id=self.id), html.Hr(), html.H2("Something")], className="inner"
        )


class Y(VizroBaseModel):
    # Higher-level container model.
    children: List[X]

    def build(self):
        return html.Div([child.build() for child in self.children], id=self.id, className="container")


def test_X_build():
    # Test for low-level contents: compare the whole component tree.
    # Sometimes setting keys_to_strip={"className"} is useful here.
    result = X(id="x", text="Hello world").build()
    expected = html.Div(
        [html.H1("Heading"), html.P("Hello world", id="x"), html.Hr(), html.H2("Something")], className="inner"
    )
    assert_component_equal(result, expected)


def test_Y_build():
    # Test for higher-level container.
    many_x = [X(text="Hello world") for _ in range(4)]
    result = Y(id="y", children=many_x).build()
    # We don't want to compare the whole component tree here. Instead compare the bit that's specifically in Y and
    # ignore the children:
    assert_component_equal(result, html.Div(id="y", className="container"), keys_to_strip={"children"})
    # And also compare the "interface" between X and Y. Use STRIP_ALL to not look at any properties of the html.Div.
    # This is basically the same as doing:
    # assert all(isinstance(child, html.Div) for child in result.children) and len(result.children) == 4
    assert_component_equal(result.children, [html.Div()] * 4, keys_to_strip=STRIP_ALL)



from typing import List, Literal

try:
    from pydantic.v1 import Field, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, validator

from vizro._constants import PARAMETER_ACTION_PREFIX
from vizro.actions import _parameter
from vizro.managers import model_manager
from vizro.models import Action, VizroBaseModel
from vizro.models._components.form import (
    Checklist,
    Dropdown,
    RadioItems,
    RangeSlider,
    Slider,
)
from vizro.models._models_utils import _log_call
from vizro.models.types import SelectorType


class Parameter(VizroBaseModel):
    """Alter the arguments supplied to any `targets` on the [`Page`][vizro.models.Page].

    Examples:
        >>> print(repr(Parameter(
        >>>    targets=["scatter.x"], selector=Slider(min=0, max=1, default=0.8, title="Bubble opacity"))))

    Args:
        type (Literal["parameter"]): Defaults to `"parameter"`.
        targets (List[str]): Targets in the form of `<target_component>.<target_argument>`.
        selector (SelectorType): See [SelectorType][vizro.models.types.SelectorType]. Converts selector value
            `"NONE"` into `None` to allow optional parameters.

    Raises:
        ValueError: If targets are invalid and not of the form `<target_component>.<target_argument>`.

    """

    type: Literal["parameter"] = "parameter"
    targets: List[str] = Field(..., description="Targets in the form of `<target_component>.<target_argument>`.")
    selector: SelectorType

    @validator("targets", each_item=True)
    def check_dot_notation(cls, target):
        if "." not in target:
            raise ValueError(
                f"Invalid target {target}. Targets must be supplied in the from of "
                "<target_component>.<target_argument>"
            )
        return target

    @validator("targets", each_item=True)
    def check_target_present(cls, target):
        target_id = target.split(".")[0]
        if target_id not in model_manager:
            raise ValueError(f"Target {target_id} not found in model_manager.")
        return target

    @validator("targets")
    def check_duplicate_parameter_target(cls, targets):
        all_targets = targets.copy()
        for _, param in model_manager._items_with_type(Parameter):
            all_targets.extend(param.targets)
        duplicate_targets = {item for item in all_targets if all_targets.count(item) > 1}
        if duplicate_targets:
            raise ValueError(f"Duplicate parameter targets {duplicate_targets} found.")
        return targets

    @_log_call
    def pre_build(self):
        self._set_slider_values()
        self._set_categorical_selectors_options()
        self._set_selector()
        self._set_actions()

    @_log_call
    def build(self):
        return self.selector.build()

    def _set_slider_values(self):
        if isinstance(self.selector, (Slider, RangeSlider)):
            if self.selector.min is None or self.selector.max is None:
                raise TypeError(
                    f"{self.selector.type} requires the arguments 'min' and 'max' when used within Parameter."
                )

    def _set_categorical_selectors_options(self):
        if isinstance(self.selector, (Checklist, Dropdown, RadioItems)) and not self.selector.options:
            raise TypeError(f"{self.selector.type} requires the argument 'options' when used within Parameter.")

    def _set_selector(self):
        if not self.selector.title:
            self.selector.title = ", ".join({target.rsplit(".")[-1] for target in self.targets})

    def _set_actions(self):
        if not self.selector.actions:
            self.selector.actions = [
                Action(
                    id=f"{PARAMETER_ACTION_PREFIX}_{self.id}",
                    function=_parameter(
                        targets=self.targets,
                    ),
                )
            ]



import logging

from vizro.managers import data_manager

logger = logging.getLogger(__name__)


def _process_callable_data_frame(captured_callable, values):
    data_frame = captured_callable["data_frame"]

    # Enable running e.g. px.scatter("iris") from the Python API and specification of "data_frame": "iris" through JSON.
    # In these cases, data already exists in the data manager and just needs to be linked to the component.
    if isinstance(data_frame, str):
        data_manager._add_component(values["id"], data_frame)
        return captured_callable

    # Standard case for px.scatter(df: pd.DataFrame).
    # Extract dataframe from the captured function and put it into the data manager.
    dataset_name = str(id(data_frame))

    logger.debug("Adding data to data manager for Figure with id %s", values["id"])
    # If the dataset already exists in the data manager then it's not a problem, it just means that we don't need
    # to duplicate it. Just log the exception for debugging purposes.
    try:
        data_manager[dataset_name] = data_frame
    except ValueError as exc:
        logger.debug(exc)

    data_manager._add_component(values["id"], dataset_name)

    # No need to keep the data in the captured function any more so remove it to save memory.
    del captured_callable["data_frame"]
    return captured_callable
